import { mkdir, rm, cp, writeFile } from 'node:fs/promises';
import { join } from 'node:path';
import { create } from 'create-svelte';
import { executeCli } from '@svelte-add/core';
import require$$0$1 from 'child_process';
import require$$0 from 'stream';
import require$$1 from 'string_decoder';
import require$$8 from 'buffer';
import { chromium } from 'playwright';
import { executeAdder, createEmptyWorkspace, populateWorkspaceDetails, createOrUpdateFiles, prompts, remoteControl } from '@svelte-add/core/internal';

const ProjectTypes = {
  Svelte_JS: "svelte-js",
  Svelte_TS: "svelte-ts",
  Kit_JS: "kit-js",
  Kit_JS_Comments: "kit-js-comments",
  Kit_TS: "kit-ts"
};
const ProjectTypesList = Object.values(ProjectTypes);
async function createProject(output, projectType) {
  await rm(output, { recursive: true, force: true });
  if (projectType.includes("kit")) {
    await create(output, {
      name: "test",
      eslint: false,
      playwright: false,
      prettier: false,
      template: "skeleton",
      types: projectType == ProjectTypes.Kit_JS_Comments ? "checkjs" : projectType == ProjectTypes.Kit_TS ? "typescript" : null,
      vitest: false
    });
  } else {
    const template = projectType == ProjectTypes.Svelte_TS ? "svelte-ts" : "svelte";
    const args = ["create", "vite@latest", projectType, "--yes", "--template", template];
    try {
      await executeCli("pnpm", args, join(output, ".."));
    } catch (error) {
      const typedError = error;
      throw new Error("Failed initializing vite project: " + typedError.message);
    }
  }
}
async function downloadProjectTemplates(output) {
  for (const template of ProjectTypesList) {
    const directory = join(output, template);
    await mkdir(directory, { recursive: true });
    await createProject(directory, template);
  }
}

async function runTests(page, adder, options) {
  const tests = {
    expectProperty: async (selector, property, expectedValue) => {
      await expectProperty(page, selector, property, expectedValue);
    },
    elementExists: async (selector) => {
      await elementExists(page, selector);
    },
    click: async (selector, path) => {
      await click(page, selector, path);
    },
    expectUrlPath: (path) => {
      expectUrlPath(page, path);
    }
  };
  await executeAdderTests(adder, tests, options);
}
async function executeAdderTests(adder, testMethods, options) {
  if (!adder.tests || adder.tests.tests.length == 0) throw new Error(`Cannot test adder without tests!`);
  for (const test of adder.tests.tests) {
    if (test.condition && !test.condition(options)) continue;
    await test.run(testMethods);
  }
}
async function elementExists(page, selector) {
  const elementToCheck = await page.$(selector);
  if (!elementToCheck) {
    throw new Error("No element found for selector " + selector);
  }
  return elementToCheck;
}
async function click(page, selector, path) {
  await elementExists(page, selector);
  await page.click(selector);
  if (path) {
    await page.waitForURL((url) => url.pathname === path);
  }
}
function expectUrlPath(page, path) {
  const url = new URL(page.url());
  if (url.pathname !== path) {
    throw new Error(`Found path ${url.pathname} but expected ${path}!`);
  }
}
async function expectProperty(page, selector, property, expectedValue) {
  const elementToCheck = await elementExists(page, selector);
  const computedStyle = await page.evaluate(([element, pV]) => window.getComputedStyle(element).getPropertyValue(pV), [
    elementToCheck,
    property
  ]);
  if (computedStyle !== expectedValue) {
    throw new Error(`Expected '${expectedValue}' but got '${computedStyle}' for selector '${selector}'`);
  }
  return computedStyle;
}

var IDX=256, HEX=[], SIZE=256, BUFFER;
while (IDX--) HEX[IDX] = (IDX + 256).toString(16).substring(1);

function uid(len) {
	var i=0, tmp=(11);
	if (!BUFFER || ((IDX + tmp) > SIZE*2)) {
		for (BUFFER='',IDX=0; i < SIZE; i++) {
			BUFFER += HEX[Math.random() * 256 | 0];
		}
	}

	return BUFFER.substring(IDX, IDX++ + tmp);
}

var commonjsGlobal = typeof globalThis !== 'undefined' ? globalThis : typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};

function getDefaultExportFromCjs (x) {
	return x && x.__esModule && Object.prototype.hasOwnProperty.call(x, 'default') ? x['default'] : x;
}

var eventStream = {};

var through$1 = {exports: {}};

(function (module, exports) {
	var Stream = require$$0;

	// through
	//
	// a stream that does nothing but re-emit the input.
	// useful for aggregating a series of changing but not ending streams into one stream)

	module.exports = through;
	through.through = through;

	//create a readable writable stream.

	function through (write, end, opts) {
	  write = write || function (data) { this.queue(data); };
	  end = end || function () { this.queue(null); };

	  var ended = false, destroyed = false, buffer = [], _ended = false;
	  var stream = new Stream();
	  stream.readable = stream.writable = true;
	  stream.paused = false;

	//  stream.autoPause   = !(opts && opts.autoPause   === false)
	  stream.autoDestroy = !(opts && opts.autoDestroy === false);

	  stream.write = function (data) {
	    write.call(this, data);
	    return !stream.paused
	  };

	  function drain() {
	    while(buffer.length && !stream.paused) {
	      var data = buffer.shift();
	      if(null === data)
	        return stream.emit('end')
	      else
	        stream.emit('data', data);
	    }
	  }

	  stream.queue = stream.push = function (data) {
	//    console.error(ended)
	    if(_ended) return stream
	    if(data === null) _ended = true;
	    buffer.push(data);
	    drain();
	    return stream
	  };

	  //this will be registered as the first 'end' listener
	  //must call destroy next tick, to make sure we're after any
	  //stream piped from here.
	  //this is only a problem if end is not emitted synchronously.
	  //a nicer way to do this is to make sure this is the last listener for 'end'

	  stream.on('end', function () {
	    stream.readable = false;
	    if(!stream.writable && stream.autoDestroy)
	      process.nextTick(function () {
	        stream.destroy();
	      });
	  });

	  function _end () {
	    stream.writable = false;
	    end.call(stream);
	    if(!stream.readable && stream.autoDestroy)
	      stream.destroy();
	  }

	  stream.end = function (data) {
	    if(ended) return
	    ended = true;
	    if(arguments.length) stream.write(data);
	    _end(); // will emit or queue
	    return stream
	  };

	  stream.destroy = function () {
	    if(destroyed) return
	    destroyed = true;
	    ended = true;
	    buffer.length = 0;
	    stream.writable = stream.readable = false;
	    stream.emit('close');
	    return stream
	  };

	  stream.pause = function () {
	    if(stream.paused) return
	    stream.paused = true;
	    return stream
	  };

	  stream.resume = function () {
	    if(stream.paused) {
	      stream.paused = false;
	      stream.emit('resume');
	    }
	    drain();
	    //may have become paused again,
	    //as drain emits 'data'.
	    if(!stream.paused)
	      stream.emit('drain');
	    return stream
	  };
	  return stream
	} 
} (through$1));

var throughExports = through$1.exports;

var Stream$2 = require$$0;

// from
//
// a stream that reads from an source.
// source may be an array, or a function.
// from handles pause behaviour for you.

var from =
function from (source) {
  if(Array.isArray(source)) {
		var source_index = 0, source_len = source.length;
    return from (function (i) {
      if(source_index < source_len)
        this.emit('data', source[source_index++]);
      else
        this.emit('end');
      return true
    })
  }
  var s = new Stream$2(), i = 0;
  s.ended = false;
  s.started = false;
  s.readable = true;
  s.writable = false;
  s.paused = false;
  s.ended = false;
  s.pause = function () {
    s.started = true;
    s.paused = true;
  };
  function next () {
    s.started = true;
    if(s.ended) return
    while(!s.ended && !s.paused && source.call(s, i++, function () {
      if(!s.ended && !s.paused)
          process.nextTick(next);
    }))
      ;
  }
  s.resume = function () {
    s.started = true;
    s.paused = false;
    next();
  };
  s.on('end', function () {
    s.ended = true;
    s.readable = false;
    process.nextTick(s.destroy);
  });
  s.destroy = function () {
    s.ended = true;
    s.emit('close'); 
  };
  /*
    by default, the stream will start emitting at nextTick
    if you want, you can pause it, after pipeing.
    you can also resume before next tick, and that will also
    work.
  */
  process.nextTick(function () {
    if(!s.started) s.resume();
  });
  return s
};

var Stream$1 = require$$0;
var writeMethods = ["write", "end", "destroy"];
var readMethods = ["resume", "pause"];
var readEvents = ["data", "close"];
var slice = Array.prototype.slice;

var duplexer$1 = duplex;

function forEach (arr, fn) {
    if (arr.forEach) {
        return arr.forEach(fn)
    }

    for (var i = 0; i < arr.length; i++) {
        fn(arr[i], i);
    }
}

function duplex(writer, reader) {
    var stream = new Stream$1();
    var ended = false;

    forEach(writeMethods, proxyWriter);

    forEach(readMethods, proxyReader);

    forEach(readEvents, proxyStream);

    reader.on("end", handleEnd);

    writer.on("drain", function() {
      stream.emit("drain");
    });

    writer.on("error", reemit);
    reader.on("error", reemit);

    stream.writable = writer.writable;
    stream.readable = reader.readable;

    return stream

    function proxyWriter(methodName) {
        stream[methodName] = method;

        function method() {
            return writer[methodName].apply(writer, arguments)
        }
    }

    function proxyReader(methodName) {
        stream[methodName] = method;

        function method() {
            stream.emit(methodName);
            var func = reader[methodName];
            if (func) {
                return func.apply(reader, arguments)
            }
            reader.emit(methodName);
        }
    }

    function proxyStream(methodName) {
        reader.on(methodName, reemit);

        function reemit() {
            var args = slice.call(arguments);
            args.unshift(methodName);
            stream.emit.apply(stream, args);
        }
    }

    function handleEnd() {
        if (ended) {
            return
        }
        ended = true;
        var args = slice.call(arguments);
        args.unshift("end");
        stream.emit.apply(stream, args);
    }

    function reemit(err) {
        stream.emit("error", err);
    }
}

//filter will reemit the data if cb(err,pass) pass is truthy

// reduce is more tricky
// maybe we want to group the reductions or emit progress updates occasionally
// the most basic reduce just emits one 'data' event after it has recieved 'end'


var Stream = require$$0.Stream;


//create an event stream and apply function to each .write
//emitting each response as data
//unless it's an empty callback

var mapStream = function (mapper, opts) {

  var stream = new Stream()
    , self = this
    , inputs = 0
    , outputs = 0
    , ended = false
    , paused = false
    , destroyed = false
    , lastWritten = 0
    , inNext = false;

  this.opts = opts || {};
  var errorEventName = this.opts.failures ? 'failure' : 'error';

  // Items that are not ready to be written yet (because they would come out of
  // order) get stuck in a queue for later.
  var writeQueue = {};

  stream.writable = true;
  stream.readable = true;

  function queueData (data, number) {
    var nextToWrite = lastWritten + 1;

    if (number === nextToWrite) {
      // If it's next, and its not undefined write it
      if (data !== undefined) {
        stream.emit.apply(stream, ['data', data]);
      }
      lastWritten ++;
      nextToWrite ++;
    } else {
      // Otherwise queue it for later.
      writeQueue[number] = data;
    }

    // If the next value is in the queue, write it
    if (writeQueue.hasOwnProperty(nextToWrite)) {
      var dataToWrite = writeQueue[nextToWrite];
      delete writeQueue[nextToWrite];
      return queueData(dataToWrite, nextToWrite)
    }

    outputs ++;
    if(inputs === outputs) {
      if(paused) paused = false, stream.emit('drain'); //written all the incoming events
      if(ended) end();
    }
  }

  function next (err, data, number) {
    if(destroyed) return
    inNext = true;

    if (!err || self.opts.failures) {
      queueData(data, number);
    }

    if (err) {
      stream.emit.apply(stream, [ errorEventName, err ]);
    }

    inNext = false;
  }

  // Wrap the mapper function by calling its callback with the order number of
  // the item in the stream.
  function wrappedMapper (input, number, callback) {
    return mapper.call(null, input, function(err, data){
      callback(err, data, number);
    })
  }

  stream.write = function (data) {
    if(ended) throw new Error('map stream is not writable')
    inNext = false;
    inputs ++;

    try {
      //catch sync errors and handle them like async errors
      var written = wrappedMapper(data, inputs, next);
      paused = (written === false);
      return !paused
    } catch (err) {
      //if the callback has been called syncronously, and the error
      //has occured in an listener, throw it again.
      if(inNext)
        throw err
      next(err);
      return !paused
    }
  };

  function end (data) {
    //if end was called with args, write it, 
    ended = true; //write will emit 'end' if ended is true
    stream.writable = false;
    if (inputs == outputs) { //wait for processing 
      stream.readable = false, stream.emit('end'), stream.destroy(); 
    }
  }

  stream.end = function (data) {
    if(ended) return
    end();
  };

  stream.destroy = function () {
    ended = destroyed = true;
    stream.writable = stream.readable = paused = false;
    process.nextTick(function () {
      stream.emit('close');
    });
  };
  stream.pause = function () {
    paused = true;
  };

  stream.resume = function () {
    paused = false;
  };

  return stream
};

//through@2 handles this by default!
var pauseStream = throughExports;

//filter will reemit the data if cb(err,pass) pass is truthy

// reduce is more tricky
// maybe we want to group the reductions or emit progress updates occasionally
// the most basic reduce just emits one 'data' event after it has recieved 'end'


var through = throughExports;
var Decoder = require$$1.StringDecoder;

var split_1 = split;

//TODO pass in a function to map across the lines.

function split (matcher, mapper, options) {
  var decoder = new Decoder();
  var soFar = '';
  var maxLength = options && options.maxLength;
  if('function' === typeof matcher)
    mapper = matcher, matcher = null;
  if (!matcher)
    matcher = /\r?\n/;

  function emit(stream, piece) {
    if(mapper) {
      try {
        piece = mapper(piece);
      }
      catch (err) {
        return stream.emit('error', err)
      }
      if('undefined' !== typeof piece)
        stream.queue(piece);
    }
    else
      stream.queue(piece);
  }

  function next (stream, buffer) {
    var pieces = ((soFar != null ? soFar : '') + buffer).split(matcher);
    soFar = pieces.pop();

    if (maxLength && soFar.length > maxLength)
      stream.emit('error', new Error('maximum buffer reached'));

    for (var i = 0; i < pieces.length; i++) {
      var piece = pieces[i];
      emit(stream, piece);
    }
  }

  return through(function (b) {
    next(this, decoder.write(b));
  },
  function () {
    if(decoder.end)
      next(this, decoder.end());
    if(soFar != null)
      emit(this, soFar);
    this.queue(null);
  })
}

var duplexer = duplexer$1;

var streamCombiner = function () {

  var streams = [].slice.call(arguments)
    , first = streams[0]
    , last = streams[streams.length - 1]
    , thepipe = duplexer(first, last);

  if(streams.length == 1)
    return streams[0]
  else if (!streams.length)
    throw new Error('connect called with empty args')

  //pipe all the streams together

  function recurse (streams) {
    if(streams.length < 2)
      return
    streams[0].pipe(streams[1]);
    recurse(streams.slice(1));  
  }
  
  recurse(streams);
 
  function onerror () {
    var args = [].slice.call(arguments);
    args.unshift('error');
    thepipe.emit.apply(thepipe, args);
  }
  
  //es.duplex already reemits the error from the first and last stream.
  //add a listener for the inner streams in the pipeline.
  for(var i = 1; i < streams.length - 1; i ++)
    streams[i].on('error', onerror);

  return thepipe
};

(function (exports) {
	//filter will reemit the data if cb(err,pass) pass is truthy

	// reduce is more tricky
	// maybe we want to group the reductions or emit progress updates occasionally
	// the most basic reduce just emits one 'data' event after it has recieved 'end'

	var Stream = require$$0.Stream
	  , es = exports
	  , through = throughExports
	  , from$1 = from
	  , duplex = duplexer$1
	  , map = mapStream
	  , pause = pauseStream
	  , split = split_1
	  , pipeline = streamCombiner
	  , immediately = commonjsGlobal.setImmediate || process.nextTick;

	es.Stream = Stream; //re-export Stream from core
	es.through = through;
	es.from = from$1;
	es.duplex = duplex;
	es.map = map;
	es.pause = pause;
	es.split = split;
	es.pipeline = es.connect = es.pipe = pipeline;
	// merge / concat
	//
	// combine multiple streams into a single stream.
	// will emit end only once

	es.concat = //actually this should be called concat
	es.merge = function (/*streams...*/) {
	  var toMerge = [].slice.call(arguments);
	  if (toMerge.length === 1 && (toMerge[0] instanceof Array)) {
	    toMerge = toMerge[0]; //handle array as arguments object
	  }
	  var stream = new Stream();
	  stream.setMaxListeners(0); // allow adding more than 11 streams
	  var endCount = 0;
	  stream.writable = stream.readable = true;

	  if (toMerge.length) {
	    toMerge.forEach(function (e) {
	      e.pipe(stream, {end: false});
	      var ended = false;
	      e.on('end', function () {
	        if(ended) return
	        ended = true;
	        endCount ++;
	        if(endCount == toMerge.length)
	          stream.emit('end');
	      });
	    });
	  } else {
	    process.nextTick(function () {
	      stream.emit('end');
	    });
	  }
	  
	  stream.write = function (data) {
	    this.emit('data', data);
	  };
	  stream.destroy = function () {
	    toMerge.forEach(function (e) {
	      if(e.destroy) e.destroy();
	    });
	  };
	  return stream
	};


	// writable stream, collects all events into an array
	// and calls back when 'end' occurs
	// mainly I'm using this to test the other functions

	es.writeArray = function (done) {
	  if ('function' !== typeof done)
	    throw new Error('function writeArray (done): done must be function')

	  var a = new Stream ()
	    , array = [], isDone = false;
	  a.write = function (l) {
	    array.push(l);
	  };
	  a.end = function () {
	    isDone = true;
	    done(null, array);
	  };
	  a.writable = true;
	  a.readable = false;
	  a.destroy = function () {
	    a.writable = a.readable = false;
	    if(isDone) return
	    done(new Error('destroyed before end'), array);
	  };
	  return a
	};

	//return a Stream that reads the properties of an object
	//respecting pause() and resume()

	es.readArray = function (array) {
	  var stream = new Stream()
	    , i = 0
	    , paused = false
	    , ended = false;

	  stream.readable = true;
	  stream.writable = false;

	  if(!Array.isArray(array))
	    throw new Error('event-stream.read expects an array')

	  stream.resume = function () {
	    if(ended) return
	    paused = false;
	    var l = array.length;
	    while(i < l && !paused && !ended) {
	      stream.emit('data', array[i++]);
	    }
	    if(i == l && !ended)
	      ended = true, stream.readable = false, stream.emit('end');
	  };
	  process.nextTick(stream.resume);
	  stream.pause = function () {
	     paused = true;
	  };
	  stream.destroy = function () {
	    ended = true;
	    stream.emit('close');
	  };
	  return stream
	};

	//
	// readable (asyncFunction)
	// return a stream that calls an async function while the stream is not paused.
	//
	// the function must take: (count, callback) {...
	//

	es.readable =
	function (func, continueOnError) {
	  var stream = new Stream()
	    , i = 0
	    , paused = false
	    , ended = false
	    , reading = false;

	  stream.readable = true;
	  stream.writable = false;

	  if('function' !== typeof func)
	    throw new Error('event-stream.readable expects async function')

	  stream.on('end', function () { ended = true; });

	  function get (err, data) {

	    if(err) {
	      stream.emit('error', err);
	      if(!continueOnError) stream.emit('end');
	    } else if (arguments.length > 1)
	      stream.emit('data', data);

	    immediately(function () {
	      if(ended || paused || reading) return
	      try {
	        reading = true;
	        func.call(stream, i++, function () {
	          reading = false;
	          get.apply(null, arguments);
	        });
	      } catch (err) {
	        stream.emit('error', err);
	      }
	    });
	  }
	  stream.resume = function () {
	    paused = false;
	    get();
	  };
	  process.nextTick(get);
	  stream.pause = function () {
	     paused = true;
	  };
	  stream.destroy = function () {
	    stream.emit('end');
	    stream.emit('close');
	    ended = true;
	  };
	  return stream
	};


	//
	// map sync
	//

	es.mapSync = function (sync) {
	  return es.through(function write(data) {
	    var mappedData;
	    try {
	      mappedData = sync(data);
	    } catch (err) {
	      return this.emit('error', err)
	    }
	    if (mappedData !== undefined)
	      this.emit('data', mappedData);
	  })
	};

	//
	// log just print out what is coming through the stream, for debugging
	//

	es.log = function (name) {
	  return es.through(function (data) {
	    [].slice.call(arguments);
	    if(name) console.error(name, data);
	    else     console.error(data);
	    this.emit('data', data);
	  })
	};


	//
	// child -- pipe through a child process
	//

	es.child = function (child) {

	  return es.duplex(child.stdin, child.stdout)

	};

	//
	// parse
	//
	// must be used after es.split() to ensure that each chunk represents a line
	// source.pipe(es.split()).pipe(es.parse())

	es.parse = function (options) {
	  var emitError = !!(options ? options.error : false);
	  return es.through(function (data) {
	    var obj;
	    try {
	      if(data) //ignore empty lines
	        obj = JSON.parse(data.toString());
	    } catch (err) {
	      if (emitError)
	        return this.emit('error', err)
	      return console.error(err, 'attempting to parse:', data)
	    }
	    //ignore lines that where only whitespace.
	    if(obj !== undefined)
	      this.emit('data', obj);
	  })
	};
	//
	// stringify
	//

	es.stringify = function () {
	  var Buffer = require$$8.Buffer;
	  return es.mapSync(function (e){
	    return JSON.stringify(Buffer.isBuffer(e) ? e.toString() : e) + '\n'
	  })
	};

	//
	// replace a string within a stream.
	//
	// warn: just concatenates the string and then does str.split().join().
	// probably not optimal.
	// for smallish responses, who cares?
	// I need this for shadow-npm so it's only relatively small json files.

	es.replace = function (from, to) {
	  return es.pipeline(es.split(from), es.join(to))
	};

	//
	// join chunks with a joiner. just like Array#join
	// also accepts a callback that is passed the chunks appended together
	// this is still supported for legacy reasons.
	//

	es.join = function (str) {

	  //legacy api
	  if('function' === typeof str)
	    return es.wait(str)

	  var first = true;
	  return es.through(function (data) {
	    if(!first)
	      this.emit('data', str);
	    first = false;
	    this.emit('data', data);
	    return true
	  })
	};


	//
	// wait. callback when 'end' is emitted, with all chunks appended as string.
	//

	es.wait = function (callback) {
	  var arr = [];
	  return es.through(function (data) { arr.push(data); },
	    function () {
	      var body = Buffer.isBuffer(arr[0]) ? Buffer.concat(arr)
	        : arr.join('');
	      this.emit('data', body);
	      this.emit('end');
	      if(callback) callback(null, body);
	    })
	};

	es.pipeable = function () {
	  throw new Error('[EVENT-STREAM] es.pipeable is deprecated')
	}; 
} (eventStream));

var spawn = require$$0$1.spawn,
    es    = eventStream;

var psTree$1 = function childrenOfPid(pid, callback) {
  var headers = null;

  if (typeof callback !== 'function') {
    throw new Error('childrenOfPid(pid, callback) expects callback');
  }

  if (typeof pid === 'number') {
    pid = pid.toString();
  }

  //
  // The `ps-tree` module behaves differently on *nix vs. Windows
  // by spawning different programs and parsing their output.
  //
  // Linux:
  // 1. " <defunct> " need to be striped
  // ```bash
  // $ ps -A -o comm,ppid,pid,stat
  // COMMAND          PPID   PID STAT
  // bbsd             2899 16958 Ss
  // watch <defunct>  1914 16964 Z
  // ps              20688 16965 R+
  // ```
  //
  // Win32:
  // 1. wmic PROCESS WHERE ParentProcessId=4604 GET Name,ParentProcessId,ProcessId,Status)
  // 2. The order of head columns is fixed
  // ```shell
  // > wmic PROCESS GET Name,ProcessId,ParentProcessId,Status
  // Name                          ParentProcessId  ProcessId   Status
  // System Idle Process           0                0
  // System                        0                4
  // smss.exe                      4                228
  // ```

  var processLister;
  if (process.platform === 'win32') {
    // See also: https://github.com/nodejs/node-v0.x-archive/issues/2318
    processLister = spawn('wmic.exe', ['PROCESS', 'GET', 'Name,ProcessId,ParentProcessId,Status']);
  } else {
    processLister = spawn('ps', ['-A', '-o', 'ppid,pid,stat,comm']);
  }

  es.connect(
    // spawn('ps', ['-A', '-o', 'ppid,pid,stat,comm']).stdout,
    processLister.stdout,
    es.split(),
    es.map(function (line, cb) { //this could parse alot of unix command output
      var columns = line.trim().split(/\s+/);
      if (!headers) {
        headers = columns;

        //
        // Rename Win32 header name, to as same as the linux, for compatible.
        //
        headers = headers.map(normalizeHeader);
        return cb();
      }

      var row = {};
      // For each header
      var h = headers.slice();
      while (h.length) {
        row[h.shift()] = h.length ? columns.shift() : columns.join(' ');
      }

      return cb(null, row);
    }),
    es.writeArray(function (err, ps) {
      var parents = {},
          children = [];

      parents[pid] = true;
      ps.forEach(function (proc) {
        if (parents[proc.PPID]) {
          parents[proc.PID] = true;
          children.push(proc);
        }
      });

      callback(null, children);
    })
  ).on('error', callback);
};

/**
 * Normalizes the given header `str` from the Windows
 * title to the *nix title.
 *
 * @param {string} str Header string to normalize
 */
function normalizeHeader(str) {
  if (process.platform !== 'win32') {
    return str;
  }

  switch (str) {
    case 'Name':
      return 'COMMAND';
    case 'ParentProcessId':
      return 'PPID';
    case 'ProcessId':
      return 'PID';
    case 'Status':
      return 'STAT';
    default:
      throw new Error('Unknown process listing header: ' + str);
  }
}

var handlePsTreeCallback$1 = function handlePsTreeCallback(
  err,
  children,
  pid,
  signal,
  options,
  callback
) {
  signal = signal || 'SIGKILL';
  options = options || {};
  callback = callback || function noop() {};
  var pollInterval = options.pollInterval || 500;
  var timeout = options.timeout;
  if (err) {
    return callback(err);
  }
  var pids = children.map(function (child) {
    return parseInt(child.PID);
  });
  pids.forEach(function (pid) {
    try {
      process.kill(pid, signal);
    } catch (error) {
      // ignore
    }
  });
  try {
    process.kill(pid, signal);
  } catch (error) {
    // don't ignore if killing the top-level process fails
    return callback(error);
  }

  pids.push(pid);

  var start = Date.now();

  function getUnterminatedProcesses() {
    var result = [];
    for (var i = 0; i < pids.length; i++) {
      var pid = pids[i];
      try {
        process.kill(pid, 0);
        // success means it's still alive
        result.push(pid);
      } catch (error) {
        // error means it's dead
      }
    }
    return result;
  }

  function poll() {
    var unterminated = getUnterminatedProcesses();
    if (!unterminated.length) {
      return callback(null);
    } else if (Date.now() + pollInterval - start > timeout) {
      return callback(new Error('timed out waiting for pids ' + unterminated.join(', ') + ' to exit'));
    } else {
      setTimeout(poll, pollInterval);
    }
  }

  setTimeout(poll, 0);
};

var psTree = psTree$1; // see: https://git.io/jBHZ
var handlePsTreeCallback = handlePsTreeCallback$1;

/**
 * terminate is an ultra-simple way to kill all the node processes
 * by providing a process pid. It finds all child processes and shuts
 * them down too, so you don't have to worry about lingering processes.
 * @param {int} pid - the Process ID you want to terminate
 * @param {string} [signal=SIGKILL] - the signal to kill the processes
 * with.
 * @param {object} [options={}] - options object
 * @param {number} [options.pollInterval=500] - interval to poll whether pid
 * and all of the child pids have been killed.
 * @param {number} [options.timeout=5000] - max time to wait for processes to
 * exit before timing out and calling back with an error.
 * @param {function=} callback - if you want to know once the
 * procesess have been terminated, supply a callback.
 * @param {Error} error - will be null if no error occured
 * @returns {void}
 */
var terminate = function terminate (pid) {
  var i = 1;
  var signal, options, callback;

  if (!pid) {
    throw new Error('No pid supplied to Terminate!');
  }

  // extract optional arguments.  Any can be omitted, but the
  // ones present have to be in the
  if (typeof arguments[i] === 'string') {
    signal = arguments[i++];
  }
  if (typeof arguments[i] === 'object' && arguments[i]) {
    options = arguments[i++];
  }
  if (typeof arguments[i] === 'function') {
    callback = arguments[i++];
  }

  psTree(pid, function (err, children) {
    handlePsTreeCallback(
      err,
      children,
      pid,
      signal,
      options,
      callback
    );
  });
};

var terminate$1 = /*@__PURE__*/getDefaultExportFromCjs(terminate);

async function startDevServer(output, command) {
  try {
    return await executeCli("pnpm", ["run", command], output, {
      onData: (data, program, resolve) => {
        const regexUnicode = /[^\x20-\xaf]+/g;
        const withoutUnicode = data.replace(regexUnicode, "");
        const regexUnicodeDigits = /\[[0-9]{1,2}m/g;
        const withoutColors = withoutUnicode.replace(regexUnicodeDigits, "");
        const regexUrl = /http:\/\/[^:\s]+:[0-9]+\//g;
        const urls = withoutColors.match(regexUrl);
        if (urls && urls.length > 0) {
          const url = urls[0];
          resolve({ url, devServer: program });
        }
      }
    });
  } catch (error) {
    const typedError = error;
    throw new Error("Failed to start dev server" + typedError.message);
  }
}
async function stopDevServer(devServer) {
  if (!devServer.pid) return;
  await forceKill(devServer);
}
async function forceKill(devServer) {
  return new Promise((resolve) => {
    if (!devServer.pid) return;
    terminate$1(devServer.pid, () => {
      resolve();
    });
  });
}

async function startBrowser(url, headless) {
  const browser = await chromium.launch({ headless });
  const page = await browser.newPage();
  await page.goto(url, { timeout: 6e4 });
  await page.waitForLoadState("networkidle");
  await page.emulateMedia({ colorScheme: "light" });
  return { browser, page };
}
async function stopBrowser(browser, page) {
  await page.close();
  await browser.close();
}

const templatesDirectory = "templates";
function getTemplatesDirectory(options) {
  return join(options.outputDirectory, templatesDirectory);
}
async function installDependencies(output) {
  try {
    await executeCli("pnpm", ["install", "--ignore-workspace"], output, { stdio: "pipe" });
  } catch (error) {
    const typedError = error;
    throw new Error("unable to install dependencies: " + typedError.message);
  }
}
async function prepareWorkspaceWithTemplate(output, template, templatesOutputDirectory) {
  const templateDirectory = join(templatesOutputDirectory, template);
  await mkdir(output, { recursive: true });
  await cp(templateDirectory, output, { recursive: true });
  return output;
}
async function saveOptionsFile(workingDirectory, options) {
  const json = JSON.stringify(options);
  await writeFile(join(workingDirectory, "options.json"), json);
}

async function runAdder(adder, workingDirectory, optionValues) {
  const adderOptions = {};
  adderOptions[adder.config.metadata.id] = optionValues;
  const remoteControlOptions = { workingDirectory, isTesting: true, adderOptions };
  await executeAdder(
    { config: adder.config, checks: adder.checks },
    { name: "@svelte-add/testing-library", version: "latest" },
    remoteControlOptions
  );
  const workspace = createEmptyWorkspace();
  workspace.cwd = workingDirectory;
  workspace.options = optionValues;
  await populateWorkspaceDetails(workspace, workingDirectory);
  await createOrUpdateFiles(adder.tests?.files ?? [], workspace);
}

var e={maxInProgress:5,failFast:!1,nextCheck:function(e,n){return Promise.resolve(e.amountStarted<n.length)},ignoreIsFunctionCheck:!1};function n(n,t){return new Promise((function(s,o){var r=Object.assign({},e,t),a={lastCompletedIndex:-1,amountDone:0,amountStarted:0,amountResolved:0,amountRejected:0,amountNextCheckFalsey:0,rejectedIndexes:[],resolvedIndexes:[],nextCheckFalseyIndexes:[],taskResults:[]};if(0===n.length)return s(a);for(var u=!1,c=0,i=function(e,n){if(a.taskResults[n]=e,a.rejectedIndexes.push(n),a.amountRejected++,!0===r.failFast)return a.lastCompletedIndex=n,u=!0,o(a);l(n);},l=function(e){if(!0!==u){if(a.amountDone++,a.lastCompletedIndex=e,"function"==typeof r.progressCallback&&r.progressCallback(a),a.amountDone===n.length)return s(a);c<n.length&&f(c++);}},f=function(e){r.nextCheck(a,n).then((function(t){!0===t?function(e){if(a.amountStarted++,"function"==typeof n[e])try{n[e]().then((function(n){a.taskResults[e]=n,a.resolvedIndexes.push(e),a.amountResolved++,l(e);}),(function(n){i(n,e);}));}catch(n){i(n,e);}else {if(!0!==r.ignoreIsFunctionCheck)return u=!0,o(new Error("tasks["+e+"]: "+n[e]+", is supposed to be of type function"));a.taskResults[e]=n[e],a.resolvedIndexes.push(e),a.amountResolved++,l(e);}}(e):(a.amountNextCheckFalsey++,a.nextCheckFalseyIndexes.push(e),l(e));}),o);},d=0;d<Math.min(r.maxInProgress,n.length);d++)f(c++);}))}

function generateTestCases(adders) {
  const testCases = /* @__PURE__ */ new Map();
  for (const adder of adders) {
    const adderTestCases = [];
    testCases.set(adder.config.metadata.id, adderTestCases);
    for (const template of ProjectTypesList) {
      const runSynchronously = adder.tests?.runSynchronously ?? false;
      const environments = adder.config.metadata.environments;
      if (!environments.kit && template.includes("kit") || !environments.svelte && template.includes("svelte")) {
        continue;
      }
      const optionsList = adder.tests?.optionValues;
      if (optionsList && optionsList.length > 0) {
        for (const options of optionsList) {
          adderTestCases.push({ adder, template, options, runSynchronously });
        }
      } else {
        const options = {};
        adderTestCases.push({ adder, template, options, runSynchronously });
      }
    }
  }
  return testCases;
}
async function runAdderTests(template, adder, options, testOptions) {
  if (!adder.tests)
    throw new Error(
      "The adder is not exporting any tests. Please make sure to properly define your tests while calling `defineAdder`"
    );
  const output = join(testOptions.outputDirectory, adder.config.metadata.id, template, uid());
  await mkdir(output, { recursive: true });
  const workingDirectory = await prepareWorkspaceWithTemplate(output, template, getTemplatesDirectory(testOptions));
  await saveOptionsFile(workingDirectory, options);
  await runAdder(adder, workingDirectory, options);
  await installDependencies(workingDirectory);
  const { url, devServer } = await startDevServer(workingDirectory, adder.tests.command ?? "dev");
  const { browser, page } = await startBrowser(url, testOptions.headless);
  try {
    const errorOcurred = await page.$("vite-error-overlay");
    if (errorOcurred) throw new Error("Dev server failed to start correctly. Vite errors present");
    if (testOptions.pauseExecutionAfterBrowser) {
      await prompts.textPrompt("Browser opened! Press any key to continue!");
    }
    await runTests(page, adder, options);
  } finally {
    await stopBrowser(browser, page);
    await stopDevServer(devServer);
  }
}
async function runTestCases(testCases, testOptions) {
  const asyncTasks = [];
  const syncTasks = [];
  const asyncTestCaseInputs = [];
  const syncTestCaseInputs = [];
  for (const values of testCases.values()) {
    for (const testCase of values) {
      const taskExecutor = async () => {
        try {
          await runAdderTests(testCase.template, testCase.adder, testCase.options, testOptions);
        } catch (e) {
          const error = e;
          const adderError = {
            name: "AdderError",
            adder: testCase.adder.config.metadata.id,
            template: testCase.template,
            message: error.message
          };
          throw adderError;
        }
      };
      if (testCase.runSynchronously) {
        syncTasks.push(taskExecutor);
        syncTestCaseInputs.push(testCase);
      } else {
        asyncTasks.push(taskExecutor);
        asyncTestCaseInputs.push(testCase);
      }
    }
  }
  let testProgressCount = 0;
  const overallTaskCount = asyncTasks.length + syncTasks.length;
  const parallelTasks = testOptions.pauseExecutionAfterBrowser ? 1 : ProjectTypesList.length;
  const allAsyncResults = await n(asyncTasks, {
    failFast: false,
    maxInProgress: parallelTasks,
    progressCallback: (result) => {
      testProgressCount++;
      logTestProgress(
        testProgressCount,
        overallTaskCount,
        result.amountResolved,
        result.amountRejected,
        asyncTestCaseInputs[result.lastCompletedIndex]
      );
    }
  });
  const allSyncResults = await n(syncTasks, {
    failFast: false,
    maxInProgress: 1,
    progressCallback: (result) => {
      testProgressCount++;
      logTestProgress(
        testProgressCount,
        overallTaskCount,
        allAsyncResults.amountResolved + result.amountResolved,
        allAsyncResults.amountRejected + result.amountRejected,
        syncTestCaseInputs[result.lastCompletedIndex]
      );
    }
  });
  const rejectedAsyncPromisesResult = allAsyncResults.rejectedIndexes.map(
    (x) => allAsyncResults.taskResults[x]
  );
  const rejectedSyncPromisesResult = allSyncResults.rejectedIndexes.map(
    (x) => allSyncResults.taskResults[x]
  );
  const rejectedPromisesResult = [...rejectedAsyncPromisesResult, ...rejectedSyncPromisesResult];
  for (const error of rejectedPromisesResult) {
    console.log(`${error.adder} (${error.template}): ${error.message}`);
  }
  if (rejectedPromisesResult.length > 0) {
    console.log("At least one test failed. Exiting.");
    process.exit(1);
  }
  if (testProgressCount != overallTaskCount) {
    console.log(
      `Number of executed tests (${testProgressCount.toString()}) does not match number of expected tests (${overallTaskCount.toString()}). Tests failed!`
    );
    process.exit(1);
  }
}
function logTestProgress(current, total, success, failed, testCaseInput) {
  const length = total.toString().length;
  const zeroPad = (num) => String(num).padStart(length, "0");
  console.log(
    `Total: ${zeroPad(current)} / ${total.toString()} Success: ${zeroPad(success)} Failed: ${zeroPad(failed)} (${testCaseInput.adder.config.metadata.id} / ${testCaseInput.template})`
  );
}

async function testAdder(adder, options) {
  await testAdders([adder], options);
}
async function testAdders(adders, options) {
  await prepareTests(options);
  remoteControl.enable();
  await executeTests(adders, options);
  remoteControl.disable();
}
async function executeTests(adders, options) {
  console.log("generating test cases");
  const testCases = generateTestCases(adders);
  console.log("start testing");
  await runTestCases(testCases, options);
}
async function prepareTests(options) {
  console.log("deleting old files");
  await rm(options.outputDirectory, { recursive: true, force: true });
  console.log("downloading project templates");
  const templatesOutputDirectory = getTemplatesDirectory(options);
  await downloadProjectTemplates(templatesOutputDirectory);
}

export { executeTests, testAdder, testAdders };
//# sourceMappingURL=index.js.map
